import gurobi.*;
import gurobi.GRB.DoubleAttr;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.poi.hssf.usermodel.HSSFWorkbook;

import java.util.Date;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;

/*
 * THIS CLASS CONTAINS THE MASTER LEVEL FRAMEWORK OF THE ADAPTIVE LARGE NEIGHBORHOOD SEARCH
 */

public class AdaptiveLargeNeighborhoodSearch {
		Mip mip;
		GRBModel model;
		Problem problem;
		int maxIterations;
		int iter;
		ArrayList<Iteration> allIterations;
		ArrayList<Iteration> bestIterations;
		long maxTimer;
		Solution solution;
		int currentInstance;
		
		// Constructor of ALNS object
		public AdaptiveLargeNeighborhoodSearch(Mip mip, Problem problem, int currentInstance){
			this.mip = mip;
			this.problem = problem;
			this.maxIterations = GlobalConstants.maximumNumberOfIterations;
			this.allIterations = new ArrayList<Iteration>();
			this.maxTimer = GlobalConstants.maxTimer;
			this.bestIterations = new ArrayList<Iteration>();
			this.currentInstance = currentInstance;
			
			// Build initial solution
			this.solution = new Solution(problem);
		}
		

		
		/*
		 * ------------------------------ METHOD TO SOLVE AN ADAPTIVE LARGE NEIGHBORHOOD SEARCH ------------------------------ 
		 */
		
		public void solveALNS(){

				// START TIMER
				long startTime = System.currentTimeMillis();
				long elapsedTime = 0L;
				int iter = 0;
				
				
					// START ADAPTIVE LARGE NEIGHBORHOOD SEARCH			
					while (iter < this.maxIterations && elapsedTime < this.maxTimer){
						
						// CREATE NEW ITERATION
						System.out.println("ITERATION: " + (iter + 1));
						Iteration currentIteration = new Iteration(iter, elapsedTime, this.maxIterations, this.solution);
						this.allIterations.add(currentIteration);
						
						
						// CALL DESTROY OPERATION
						DestroyOperator.destroy(this.solution, currentIteration);
						
						
						// CALL REPAIR OPERATION
						RepairOperator.repair(this.solution, currentIteration, this.mip);
						
						
						// CALL LOCAL SEARCH OPERATION
						LocalSearch.localSearch(this.solution, currentIteration);
						
						
						//TEST
						/*
						boolean possibleSolution = true;
						possibleSolution = RepairOperator.callMipSolver(solution, mip);
						solution.updateSolutionObject(mip);
						if (!possibleSolution){
							System.exit(1);
						}
						*/
						
						
						// EVALUATE ITERATION
						currentIteration.addSolution(this.solution);
						currentIteration.evaluate(this.allIterations, this.solution, this.bestIterations);
						
						
						// UPDATE STOPPING CRITERAS
						elapsedTime = (new Date()).getTime() - startTime;
						iter++;
						
					}
						
				// POST-SEARCH EVALUATION. START BY COLLECTING BEST SOLUTION
				Iteration bestIteration = this.findBestIteration(bestIterations);
				
				// Check if the current solution is a "quality-solution"
				
				
				// PRINT BEST FOUND SOLUTION TO SCREEN
				System.out.println("BEST SOLUTION FOUND WAS FOUND IN ITERATION " + bestIteration.iterationNumber + ".");
				System.out.println("BEST SOLUTION FOUND: " + bestIteration.objectiveValue + ".");
				System.out.println("BEST SOLUTION WAS FOUND AFTER " + (bestIteration.iterationTime) + " MILLISECONDS");
				System.out.println("NUMBER OF DUMMIES: " + bestIteration.numberOfDummies);
				
				
				// WRITE TO FILE
				this.writeALNStoFile(bestIteration);
				
		}
		
		
		
		
		
		/*
		 * -------------------------------- METHOD TO WRITE ALNS RESULT TO FILE ----------------------------------------------
		 */
		
		public void writeALNStoFile(Iteration bestIteration){
			
			Writer writer = null;
			
			try {
			    writer = new BufferedWriter(new OutputStreamWriter(
			          new FileOutputStream("../Output/ALNS3_OUTPUT_" + TestInstances.testInstances.get(this.currentInstance) + ".txt"), "utf-8"));
			    writer.write("OUTPUT FROM SOLVING " + this.problem.instanceName + " WITH ALNS");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("NUMBER OF ITERATIONS: " + this.allIterations.size());
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("TIME ELAPSED: " + (allIterations.get(this.allIterations.size()-1).iterationTime)/1000 + " SECONDS");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("THE FIRST ITERATION WITH " + bestIteration.numberOfDummies + " DUMMIES WERE FOUND IN ITERATION " + this.bestIterations.get(0).iterationNumber + " AFTER " + this.bestIterations.get(0).iterationTime + " MILLISECONDS.");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("BEST SOLUTION FOUND WAS FOUND IN ITERATION " + bestIteration.iterationNumber + " AFTER " + bestIteration.iterationTime + " MILLISECONDS.");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("BEST SOLUTION FOUND: " + bestIteration.objectiveValue + ".");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write(System.getProperty( "line.separator" ));
			    int quality = isIterationQualitySolution(bestIterations);
			    if (quality >= 0){
			    	writer.write("FIRST QUALITY SOLUTION WAS FOUND IN ITERATION " + this.bestIterations.get(quality).iterationNumber + " AFTER " + this.bestIterations.get(quality).iterationTime + " MILLISECONDS.");
			    }
			    else{
			    	writer.write("NO QUALITY SOLUTIONS WAS FOUND");
			    }
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("VOYAGES COVERED BY DUMMY VARIABLES:" );
			    writer.write(System.getProperty( "line.separator" ));
					for (int i = 0; i < this.problem.positions.size(); i++){
						if (bestIteration.dummyPositions[i] > 0.5){
					    writer.write("" + this.problem.positions.get(i).id);
					    writer.write(System.getProperty( "line.separator" ));
						}
					}
			    writer.write("-------------------------------------------------------");
				  writer.write(System.getProperty( "line.separator" ));
			    writer.write("VOYAGE ASSIGNMENTS: ");
			    writer.write(System.getProperty( "line.separator" ));
					for (int v = 0; v < this.solution.vessels.size();v++){
				    writer.write("VESSEL " + this.solution.vessels.get(v).code + " SAILS:");
				    writer.write(System.getProperty( "line.separator" ));
						for (int i = 0; i < this.solution.vessels.get(v).assignedPositions.size(); i++){
							writer.write("" + this.solution.vessels.get(v).assignedPositions.get(i).id);
						   writer.write(System.getProperty( "line.separator" ));
						}
					}	
			    writer.write("-------------------------------------------------------");
			    
			    
			    writer.write(System.getProperty( "line.separator" ));
					writer.write("WEIGHT DEVELOPMENTS: ");
					writer.write(System.getProperty( "line.separator" ));
					
					/*
					for (int j = 0; j < (GlobalConstants.maximumNumberOfIterations / GlobalConstants.segmentSize); j++){
						writer.write((j+1) + " " + solution.destroyWeights[0][j] + " " + solution.destroyWeights[1][j] + " " + solution.destroyWeights[2][j] + " " + solution.destroyWeights[3][j] + " " + solution.destroyWeights[4][j] + " ");
					   writer.write(System.getProperty( "line.separator" ));
					}
					*/
					
					
					String segmentWeights = "";
					for (int j = 0; j < (GlobalConstants.maximumNumberOfIterations / GlobalConstants.segmentSize); j++){
						segmentWeights += "WEIGHTS IN SEGMENT " + j + " = ";
						for (int d = 0; d < GlobalConstants.numberOfDestroyOperators; d++){
							for (int r = 0; r < GlobalConstants.numberOfRepairOperators; r++){		
								String weight = String.format("%.2f", this.solution.weights[d][r][j]);		
								segmentWeights += "[" + weight + "]";
							}
						}
						writer.write(segmentWeights);
					  writer.write(System.getProperty( "line.separator" ));
						segmentWeights = "";
					}
					
					
					
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
					writer.write("OBJECTIVE VALUE DEVELOPMENT: ");
			    writer.write(System.getProperty( "line.separator" ));
			    int maxBestIterations;
			    if (bestIterations.size() > GlobalConstants.maxNumberOfBestIterations){
			    	maxBestIterations = GlobalConstants.maxNumberOfBestIterations;
			    }else{
			    	maxBestIterations = bestIterations.size();
			    }
			    
					for (int k = 0; k < maxBestIterations; k++){
						writer.write("ITERATION: " + this.bestIterations.get(k).iterationNumber + ", TIME: " + ((this.bestIterations.get(k).iterationTime)/(1000)) + ", OBJ.VALUE: " + this.bestIterations.get(k).objectiveValue);
						writer.write(System.getProperty( "line.separator" ));
					}
					
					
					
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
					writer.write("NUMBER OF UNSERVED VOYAGES DEVELOPMENT: ");
				  writer.write(System.getProperty( "line.separator" ));
					for (int k = 0; k < this.allIterations.size(); k++){
						if (((k+1) % 5) == 0){
							writer.write(" " + this.allIterations.get(k).numberOfDummies);
							writer.write(System.getProperty( "line.separator" ));
						}
					}
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));

			    		
			} catch (IOException ex) {
			  // report
			} finally {
			   try {writer.close();} catch (Exception ex) {}
			}
			

		}
		
		
		
		/*
		 * -------------- Method to retrieve the best iteration from a list of iterations -------------------
		 */
		
		public Iteration findBestIteration(ArrayList<Iteration> bestIterations){
			
			Iteration bestIteration = bestIterations.get(0);
			double bestObjValue = Double.POSITIVE_INFINITY;
			int maxIterations;
			
			// Limit the number of best iterations to check
			if (bestIterations.size() > GlobalConstants.maxNumberOfBestIterations){
				maxIterations = GlobalConstants.maxNumberOfBestIterations;
			}else{
				maxIterations = bestIterations.size();
			}
			
			
			// Iterate all best iterations found
			for (int k = 0; k < maxIterations; k++){
				
				try{

					// Send the iteration to the Solution object, and send the Solution object as a fixed LP to the MIP solver.
					bestIterations.get(k).getSolutionFromIteration(this.solution);				
					RepairOperator.solutionToModel(this.solution, this.mip);
					bestIterations.get(k).objectiveValue = this.mip.model.get(DoubleAttr.ObjVal);
				
					if (this.mip.model.get(DoubleAttr.ObjVal) < bestObjValue){
						
						// We have a new best solution
						bestIteration = bestIterations.get(k);
						bestObjValue = this.mip.model.get(DoubleAttr.ObjVal);
						bestIteration.objectiveValue = this.mip.model.get(DoubleAttr.ObjVal);
						
					}
					
				}catch(GRBException e) {
		      System.out.println("Error code: " + e.getErrorCode() + ". " +
		          e.getMessage());
		      e.printStackTrace();
				}
			}
			
			
			bestIteration.getSolutionFromIteration(this.solution);				
			return bestIteration;
		}
	
		/*
		 * -------------- Method to check if a solution is a quality solution-------------------
		 */
		public int isIterationQualitySolution (ArrayList<Iteration> bestIterations){
			for (int i = 0; i< bestIterations.size(); i++){
	    		if (((bestIterations.get(i).objectiveValue - GlobalConstants.bestBounds.get(this.currentInstance))
	    				/ (GlobalConstants.bestBounds.get(this.currentInstance)) ) <= GlobalConstants.qualityGAPs.get(this.currentInstance)){
					return i;
					
				}
			}
			return -1;
		}
}
