

import gurobi.GRB;
import gurobi.GRB.DoubleAttr;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Date;
import java.util.ArrayList;
import java.util.List;

/*
 * 
 * This class builds mixed-integer programs from problem sets, and contains methods to solve them.
 * 
 */

public class Mip {
	Problem problem;
	GRBModel model;
	GRBEnv env;

	GRBVar[][][]y;
	GRBVar[][]y_i;
	GRBVar[][]y_o;
	double startTime;
	double endTime;
	int currentInstance;
	
		/*
		 * ---------------- Constructor to build mathematical program from problem instance --------------
		 */
	
		public Mip(Problem problem){
			this.problem = problem;
			this.startTime = 0;
			this.endTime = 0;
			this.currentInstance = currentInstance;
			
			// START TIMER for whole MIP-model
			this.startTime = System.currentTimeMillis();
			
		    
			try{
				

				
				// Construct model    
				
	      GRBEnv env = new GRBEnv();
				GRBModel model = new GRBModel(env);
				model.set(GRB.StringAttr.ModelName, "fleetDeployment");
				

	      // Generate variables
	      // Vessel sailing from position to position decision variables: y[v][i][j] == 1 if vessel v sails from i to j.
	      GRBVar[][][] y = new GRBVar[problem.vessels.size()][problem.positions.size()][problem.positions.size()];
	      for (int v = 0; v < problem.vessels.size(); v++){
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		for (int j = 0; j < problem.positions.size(); j++){
	      			if (problem.vesselCanSailPositionToPosition(problem.vessels.get(v), problem.positions.get(i), problem.positions.get(j))){
	      				y[v][i][j] = model.addVar(0, 1, 0, GRB.BINARY, "y[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "][" + problem.positions.get(j).id + "]");
	      			}
	    		}
	      	}
	      }
	      this.y = y;
	      
	      // Vessel sailing from start to position decision variables: y_o[v][i] == 1 if vessel v sails from start to i.
	      GRBVar[][] y_o = new GRBVar[problem.vessels.size()][problem.positions.size()];
	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailInitialPosition(problem.vessels.get(v),problem.positions.get(i))){
	      			y_o[v][i] = model.addVar(0, 1, 0, GRB.BINARY, "y_o[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "]");
	      		}
	      	}
	      }
	     this.y_o = y_o;
	     
	      // Vessel sailing from position to artificial destination decision variables: y_d[v][i] == 1 if vessel v sails from position i to artificial destination.
	      GRBVar[][] y_d = new GRBVar[problem.vessels.size()][problem.positions.size()];
	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){ 
	      		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			y_d[v][i] = model.addVar(0, 1, 0, GRB.BINARY, "y_d[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "]");
	      		}
	      	}
	      }
	      
	      
	      // Vessel sailing from position decision variables: y[v][i] == 1 if vessel v sails position i. Connection variable.
	      GRBVar[][] y_i = new GRBVar[problem.vessels.size()][problem.positions.size()];
	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			y_i[v][i] = model.addVar(0, 1, 0, GRB.BINARY, "y_i[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "]");
	      		}
	    	}
	      }
	      this.y_i = y_i;
	      
	      // Vessel sailing directly from initial position to artificial destination, lambda, i.e. vessel is not used
	      GRBVar[] lambda = new GRBVar[problem.vessels.size()];
	      for (int v = 0; v < problem.vessels.size(); v++) {     					
	      	lambda[v] = model.addVar(0, 1, GlobalConstants.costOfNotSailing, GRB.BINARY, "lambda[" + problem.vessels.get(v).code + "]");
	      }
	      
	      
	      // Speed weights for vessel sailing from origin to position decision variables: x_Bo[v][i][s] == 1 if vessel v sails from origin to position i on speed profile s.
	      GRBVar[][][] x_Bo = new GRBVar[problem.vessels.size()][problem.positions.size()][problem.speedProfiles.size()];
	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		for (int s = 0; s < problem.speedProfiles.size(); s++){
	      			if (problem.vesselCanSailInitialPosition(problem.vessels.get(v), problem.positions.get(i))){
	      				double cost = SailingBallastStart.getStartCosts(problem.vessels.get(v).code, problem.positions.get(i).route, problem.speedProfiles.get(s).code, problem.sailingBallastStarts);
	      				x_Bo[v][i][s] = model.addVar(0, 1, cost, GRB.CONTINUOUS, "x_Bo[" + problem.vessels.get(v).code + "][" + problem.speedProfiles.get(s).code + "][" + problem.positions.get(i).id  + "]");
	      			}
	      		}
	      	}
	      }

	      
	      // Speed weights for vessel sailing position decision variables: x[v][i][s] == 1 if vessel v sails position i with speed profile s.
	      GRBVar[][][] x = new GRBVar[problem.vessels.size()][problem.positions.size()][problem.speedProfiles.size()];
	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		for (int s = 0; s < problem.speedProfiles.size(); s++){
	      			if (problem.vesselCanSailPosition(problem.vessels.get(v),problem.positions.get(i))){
	        			double cost = Sailing.getCosts(problem.vessels.get(v).code, problem.positions.get(i).route, problem.speedProfiles.get(s).code, problem.sailings);
	      				x[v][i][s] = model.addVar(0, 1, cost, GRB.CONTINUOUS, "x[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "][" + problem.speedProfiles.get(s).code  + "]");
	      			}
	      		}
	      	}
	      }
	      
	      
	      // Speed weights for vessel sailing ballast between positions decision variables: x_B[v][i][j][s] == 1 if vessel v sails from i to j on speed profile s.
	      GRBVar[][][][] x_B = new GRBVar[problem.vessels.size()][problem.positions.size()][problem.positions.size()][problem.speedProfiles.size()];
	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		for (int j = 0; j < problem.positions.size(); j++){
	      			if (problem.vesselCanSailPositionToPosition(problem.vessels.get(v), problem.positions.get(i), problem.positions.get(j))){
	      				for (int s = 0; s < problem.speedProfiles.size(); s++){
	      					double cost = SailingBallast.getCosts(problem.vessels.get(v).code, problem.positions.get(i).route, problem.positions.get(j).route, problem.speedProfiles.get(s).code, problem.sailingBallasts);
	      					x_B[v][i][j][s] = model.addVar(0, 1, cost, GRB.CONTINUOUS, "x_B[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "][" + problem.positions.get(j).id +  "]["  + problem.speedProfiles.get(s).code + "]");
	      				}
	      			}
	      		}
	      	}
	      }
	      
	      
	      // Dummy variable / chartered ships for uncovered voyages. s[i] = 1 if voyage i is not sailed by the fleet.
	      GRBVar[] dummy = new GRBVar[problem.positions.size()];
	      for (int i = 0; i < problem.positions.size(); i++){
	      	dummy[i] = model.addVar(0, 1, GlobalConstants.costOfVesselCharter, GRB.BINARY, "dummy[" + problem.positions.get(i).id + "]");
	      }      
	      
	      
	      // Start times for vessels from initial position
	      GRBVar[] t_o = new GRBVar[problem.vessels.size()];
	      for (int v = 0; v < problem.vessels.size(); v++) {
	        double dateOffset = problem.vessels.get(v).dateOffset;
	      	t_o[v] = model.addVar(dateOffset, Position.getLatestVoyageStart(problem.positions) + GlobalConstants.maxDaysDelayed, 0, GRB.CONTINUOUS, "t_o[" + problem.vessels.get(v).code + "]");
	      }
	      
	      
	      // Start times for positions
	      GRBVar[] t = new GRBVar[problem.positions.size()];
	      for (int i = 0; i < problem.positions.size(); i++) {
	        double startTime = problem.positions.get(i).startTime;
	      	t[i] = model.addVar(startTime, Position.getLatestVoyageStart(problem.positions) + GlobalConstants.maxDaysDelayed, 0, GRB.CONTINUOUS, "t[" + problem.positions.get(i).id + "]");
	      }
	      
	      
	      // Position delay times
	      GRBVar[] d = new GRBVar[problem.positions.size()];
	      for (int i = 0; i < problem.positions.size(); i++) {
	      	d[i] = model.addVar(0, GlobalConstants.maxDaysDelayed, GlobalConstants.penaltyCostPerDayDelayed, GRB.CONTINUOUS, "d[" + problem.positions.get(i).id + "]");
	      }


	      // Load variables 
	      GRBVar[][][][] l = new GRBVar[problem.positions.size()][problem.balanceCategories.size()][problem.deckCategories.size()][problem.deckCategories.size()];
	      for (int i = 0; i < problem.positions.size(); i++) {
	      	for (int b = 0; b < problem.balanceCategories.size(); b++){
	      		for (int p = 0; p < problem.deckCategories.size(); p++){
	      			for (int k = 0; k < problem.deckCategories.size(); k++){
	      				if (BalanceCategoryStandardRoute.isBalanceCategoryInRoute(problem.balanceCategories.get(b).code, problem.positions.get(i).route, problem.balanceCategoryStandardRoutes)){
	      					if (DeckCategoryToDeckCategory.ifPossibleToPut(problem.deckCategories.get(p).code, problem.deckCategories.get(k).code, problem.deckCategoryToDeckCategories)){
	      						l[i][b][p][k] = model.addVar(0, VesselSetup.getMaxCapacity(problem.vesselSetups, problem.deckCategories.get(k).code), 0, GRB.CONTINUOUS, "l[" + problem.positions.get(i).id + "][" + problem.balanceCategories.get(b).code + "][" + problem.deckCategories.get(p).code + "][" + problem.deckCategories.get(k).code + "]");
	      					}
	      				}
	      			}
	      		}	
	      	}
	      }
	      
	      
	      // Space charter variables
	      GRBVar[][][] z = new GRBVar[problem.balanceCategories.size()][problem.deckCategories.size()][problem.months.size()];
	      for (int b = 0; b < problem.balanceCategories.size(); b++) {
	      	for (int p = 0; p < problem.deckCategories.size(); p++){
	      		for (int m = 0; m < problem.months.size(); m++){
	      			if (Demand.isDeckCategoryInBalanceCategory(problem.deckCategories.get(p).code, problem.balanceCategories.get(b).code, problem.demands)){
	      				z[b][p][m] = model.addVar(0, Demand.getSpaceCharterCap(problem.balanceCategories.get(b).code, problem.months.get(m).code, problem.deckCategories.get(p).code, problem.demands), Demand.getSpaceCharterCost(problem.balanceCategories.get(b).code, problem.months.get(m).code, problem.deckCategories.get(p).code, problem.demands), GRB.CONTINUOUS, "z[" + problem.balanceCategories.get(b).code + "][" + problem.deckCategories.get(p).code + "][" + problem.months.get(m).code + "]");
	      			}
	      		}
	      	}
	      }   
				
	      
	      
	      // Update model to integrate variables
	      model.update();
	      
	      

	      // The objective is to minimize costs
	      model.set(GRB.IntAttr.ModelSense, 1);

	      
	      
	      // Set branching strategy	      
	      for (int i = 0; i < problem.positions.size(); i++){
	      	dummy[i].set(GRB.IntAttr.BranchPriority, 2);
	      	for (int v = 0; v < problem.vessels.size(); v++){
	      		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			y_i[v][i].set(GRB.IntAttr.BranchPriority, 1);
	      		}
	      	}
	      }
	      
	      
	      
	      // Generate constraints
	      
	      
	      // 4.1 
	      // Objective Function
	      // Generated directly in variable declarations
	      
	      
	      // 4.2
	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailInitialPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			GRBLinExpr sailingInitialBallastConstrLHS = new GRBLinExpr();
	      			sailingInitialBallastConstrLHS.addTerm(1.0, y_o[v][i]);
	      		
	      			GRBLinExpr sailingInitialBallastSpeedConstrRHS = new GRBLinExpr();
	      			for (int s = 0; s < problem.speedProfiles.size(); s++){
	      				sailingInitialBallastSpeedConstrRHS.addTerm(1.0, x_Bo[v][i][s]);
	      			}
	      			model.addConstr(sailingInitialBallastConstrLHS, GRB.EQUAL, sailingInitialBallastSpeedConstrRHS, "Initial ballast flow and speed connection [" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "]");
	      		}
	      	}
	      }
	      
	      
	      // 4.3
	      for (int v = 0; v < problem.vessels.size(); v++){
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			GRBLinExpr sailingSpeedConstrLHS = new GRBLinExpr();
	      			sailingSpeedConstrLHS.addTerm(1.0, y_i[v][i]);
	      		
	      			GRBLinExpr sailingSpeedConstrRHS = new GRBLinExpr();
	      			for (int s = 0; s < problem.speedProfiles.size(); s++){
	      				sailingSpeedConstrRHS.addTerm(1.0, x[v][i][s]);	
	      			}
	      		
	      			model.addConstr(sailingSpeedConstrLHS, GRB.EQUAL, sailingSpeedConstrRHS, "Sailing flow and speed connection [" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "]");
	      		}
	      	}
	      }
	      
	      
	      // 4.4
	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		for (int j = 0; j < problem.positions.size(); j++){
	      			if (problem.vesselCanSailPositionToPosition(problem.vessels.get(v), problem.positions.get(i), problem.positions.get(j))){
	      				GRBLinExpr sailingBallastConstrLHS = new GRBLinExpr();
	      				sailingBallastConstrLHS.addTerm(1.0, y[v][i][j]);
	  	
	      				GRBLinExpr sailingBallastSpeedConstrRHS = new GRBLinExpr();
	      				for (int s = 0; s < problem.speedProfiles.size(); s++){
	      					sailingBallastSpeedConstrRHS.addTerm(1.0, x_B[v][i][j][s]);
	      				}
	      				
	          		model.addConstr(sailingBallastConstrLHS, GRB.EQUAL, sailingBallastSpeedConstrRHS, "Sailing ballast flow and speed connection [" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "][" + problem.positions.get(j).id + "]");
	      			}
	      		}
	      	}
	      }
	     
	      
	      // 4.5
	      for (int i = 0; i < problem.positions.size(); i++) {
	      	if (!problem.positions.get(i).isMaintenance(problem.positions.get(i))){
	      		GRBLinExpr voyageServedConstrLHS = new GRBLinExpr();
	      		
	      		for (int v = 0; v < problem.vessels.size(); v++){
	      			if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	      				voyageServedConstrLHS.addTerm(1.0, y_i[v][i]);
	      			}
	      		}
	      		voyageServedConstrLHS.addTerm(1.0, dummy[i]);
	      		
	      		GRBLinExpr voyageServedConstRHS = new GRBLinExpr();
	      		voyageServedConstRHS.addConstant(1.0);
	      		
	      		model.addConstr(voyageServedConstrLHS, GRB.EQUAL, voyageServedConstRHS, "Voyage served [" + problem.positions.get(i).id + "]");
	      	}
	      }
	      
	      
	      // 4.6 (foreløpig tatt bort, fungerer ikke)
	      /*
	      for (int i = 0; i < problem.positions.size(); i++) {
	      	if (problem.positions.get(i).isMaintenance(problem.positions.get(i))){
	    			for (int v = 0; v <= problem.vessels.size(); v++){
	        		GRBLinExpr voyageMaintenanceServedConstrLHS = new GRBLinExpr();
	        		
	        		for (int j = 0; j <= problem.positions.size(); j++){
	      				if (i != j && (problem.positions.get(i).startTime < problem.positions.get(j).endTime)){
	      					voyageMaintenanceServedConstrLHS.addTerm(1.0, y[v][i][j]);
	      				}
	      			}
	      			voyageMaintenanceServedConstrLHS.addTerm(1.0, y_d[v][i]);
	      			
	        		GRBLinExpr voyageMaintenanceServedConstrRHS = new GRBLinExpr();
	        		voyageMaintenanceServedConstrRHS.addConstant(1.0);
	        		
	      			model.addConstr(voyageMaintenanceServedConstrLHS, GRB.EQUAL, 1, "Maintenance voyage served [" + problem.positions.get(i).id + "]");
	      		}
	      	}
	      }
	      */
	      
	      
	      // 4.7
	      for (int v = 0; v < problem.vessels.size(); v++){      	
	      	GRBLinExpr originFlowConstrLHS = new GRBLinExpr();
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailInitialPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			originFlowConstrLHS.addTerm(1.0, y_o[v][i]);
	      		}
	      	}
	  			originFlowConstrLHS.addTerm(1.0,lambda[v]);

	  			GRBLinExpr originFlowConstrRHS = new GRBLinExpr();
	  			originFlowConstrRHS.addConstant(1.0);
	  				
	  			model.addConstr(originFlowConstrLHS, GRB.EQUAL, originFlowConstrRHS, "Vessel leaves origin [" + problem.vessels.get(v).code + "]");
	      }
	      
	      
	      // 4.8
	      for (int v = 0; v < problem.vessels.size(); v++){      	
	      	GRBLinExpr destinationFlowConstrLHS = new GRBLinExpr();
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			destinationFlowConstrLHS.addTerm(1.0, y_d[v][i]);
	      		}
	      	}
	  			destinationFlowConstrLHS.addTerm(1.0,lambda[v]);

	  			GRBLinExpr destinationFlowConstrRHS = new GRBLinExpr();
	  			destinationFlowConstrRHS.addConstant(1.0);
	  				
	  			model.addConstr(destinationFlowConstrLHS, GRB.EQUAL, destinationFlowConstrRHS, "Vessel arrives destination [" + problem.vessels.get(v).code + "]");
	      }
	      
	      
	      // 4.9
	      for (int v = 0; v < problem.vessels.size(); v++){
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			GRBLinExpr yConnectBConstrLHS = new GRBLinExpr();
	      			yConnectBConstrLHS.addTerm(1.0, y_i[v][i]);
	      			
	      			GRBLinExpr yConnectBConstrRHS = new GRBLinExpr();
	      			for (int j = 0; j < problem.positions.size(); j++){
	      				if (problem.vesselCanSailPositionToPosition(problem.vessels.get(v), problem.positions.get(j), problem.positions.get(i))){
	      					yConnectBConstrRHS.addTerm(1.0, y[v][j][i]);
	      				}
	      			}
	      			if (problem.vesselCanSailInitialPosition(problem.vessels.get(v), problem.positions.get(i))){
	      				yConnectBConstrRHS.addTerm(1.0, y_o[v][i]);
	      			}
	      		
	      			model.addConstr(yConnectBConstrLHS, GRB.EQUAL, yConnectBConstrRHS, "B Connect y_vij and y_vi [" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "]");
	      		}
	      	}
	      }
	      
	      
	      // 4.10
	      for (int v = 0; v < problem.vessels.size(); v++){
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			GRBLinExpr yConnectAConstrLHS = new GRBLinExpr();
	      			yConnectAConstrLHS.addTerm(1.0, y_i[v][i]);
	      		
	      			GRBLinExpr yConnectAConstrRHS = new GRBLinExpr();
	      			for (int j = 0; j < problem.positions.size(); j++){
	      				if (problem.vesselCanSailPositionToPosition(problem.vessels.get(v), problem.positions.get(i), problem.positions.get(j))){
	      					yConnectAConstrRHS.addTerm(1.0, y[v][i][j]);
	      				}
	      			}
	      			yConnectAConstrRHS.addTerm(1.0, y_d[v][i]);
	      		
	      			model.addConstr(yConnectAConstrLHS, GRB.EQUAL, yConnectAConstrRHS, "A Connect y_vij and y_vi [" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "]");
	      		}
	      	}
	      }
	      
	      
	      // 4.11 (linearized as in Appendix)
	      for (int v = 0; v < problem.vessels.size(); v++){
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		if (problem.vesselCanSailInitialPosition(problem.vessels.get(v), problem.positions.get(i))){
	      			GRBLinExpr timeOriginConstrLHS = new GRBLinExpr();
	      			timeOriginConstrLHS.addTerm(1.0, t[i]);
	      		
	      			GRBLinExpr timeOriginConstrRHS = new GRBLinExpr();
	      			timeOriginConstrRHS.addTerm(1.0, t_o[v]);
	      			for (int s = 0; s < problem.speedProfiles.size(); s++){
	      				double sailingBallastOriginTime = SailingBallastStart.getStartDuration(problem.vessels.get(v).code, problem.positions.get(i).route, problem.speedProfiles.get(s).code, problem.sailingBallastStarts);
	      				timeOriginConstrRHS.addTerm(sailingBallastOriginTime, x_Bo[v][i][s]);
	      			}
	      			
	      			//timeOriginConstrRHS.addConstant(- problem.months.get(problem.months.size()-1).end);
	      			//timeOriginConstrRHS.addTerm(problem.months.get(problem.months.size()-1).end, y_o[v][i]);
	      			
	      			timeOriginConstrRHS.addConstant(- (Sailing.getLongestSailingDuration(problem.sailings) + Position.getLatestVoyageStart(problem.positions) + GlobalConstants.maxDaysDelayed));
	      			timeOriginConstrRHS.addTerm(Sailing.getLongestSailingDuration(problem.sailings) + Position.getLatestVoyageStart(problem.positions) + GlobalConstants.maxDaysDelayed, y_o[v][i]);
	      		
	      			model.addConstr(timeOriginConstrLHS, GRB.GREATER_EQUAL, timeOriginConstrRHS, "Time constraint [" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "]");
	      		}
	      	}
	      }

	      
	      // 4.12 (linearized as in Appendix)
	      for (int v = 0; v < problem.vessels.size(); v++){
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		for (int j = 0; j < problem.positions.size(); j++){
	      			if (problem.vesselCanSailPositionToPosition(problem.vessels.get(v), problem.positions.get(i), problem.positions.get(j))){
	          		GRBLinExpr timeConstrLHS = new GRBLinExpr();
	          		timeConstrLHS.addTerm(1.0, t[j]);
	          		    				
	          		GRBLinExpr timeConstrRHS = new GRBLinExpr();
	          		timeConstrRHS.addTerm(1.0, t[i]);
	          		for (int s = 0; s < problem.speedProfiles.size(); s++){
	          			double sailingTime = Sailing.getSailingDuration(problem.vessels.get(v).code, problem.positions.get(i).route, problem.speedProfiles.get(s).code, problem.sailings);
	          			timeConstrRHS.addTerm(sailingTime, x[v][i][s]);
	          		}
	          		for (int s = 0; s < problem.speedProfiles.size(); s++){
	          			double sailingBallastTime = SailingBallast.getSailingDuration(problem.vessels.get(v).code, problem.positions.get(i).route, problem.positions.get(j).route, problem.speedProfiles.get(s).code, problem.sailingBallasts);
	          			timeConstrRHS.addTerm(sailingBallastTime, x_B[v][i][j][s]);
	          		}
	          		timeConstrRHS.addConstant(- (Sailing.getLongestSailingDuration(problem.sailings) + Position.getLatestVoyageStart(problem.positions) + GlobalConstants.maxDaysDelayed));
	        			timeConstrRHS.addTerm(Sailing.getLongestSailingDuration(problem.sailings) + Position.getLatestVoyageStart(problem.positions) + GlobalConstants.maxDaysDelayed, y[v][i][j]);
	        			
	        			model.addConstr(timeConstrLHS, GRB.GREATER_EQUAL, timeConstrRHS, "Time constraint [" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "][" + problem.positions.get(j).id + "]");
	      			}
	      		}
	      	}
	      }
	      
	      
	      // 4.13
	      for (int i = 0; i < problem.positions.size(); i++){      	
	      	GRBLinExpr voyageDelayedConstrLHS = new GRBLinExpr();
	      	voyageDelayedConstrLHS.addTerm(1.0, d[i]);
	      
	  			GRBLinExpr voyageDelayedConstrRHS = new GRBLinExpr();
	  			voyageDelayedConstrRHS.addTerm(1.0, t[i]);
	  			voyageDelayedConstrRHS.addConstant(- problem.positions.get(i).endTime);
	  				
	  			model.addConstr(voyageDelayedConstrLHS, GRB.GREATER_EQUAL, voyageDelayedConstrRHS, "Days delayed [" + problem.positions.get(i).id + "]");
	      }
	      
	      
	      // 4.14
  	    for (int m = 0; m < problem.months.size(); m++){
  	    	for (int b = 0; b < problem.balanceCategories.size(); b++){
  	    		for (int p = 0; p < problem.deckCategories.size(); p++){
  	    			if (Demand.isDeckCategoryInBalanceCategory(problem.deckCategories.get(p).code, problem.balanceCategories.get(b).code, problem.demands)){
	            	GRBLinExpr monthlyDemandConstrLHS = new GRBLinExpr();
	            	monthlyDemandConstrLHS.addTerm(1.0, z[b][p][m]);
	            	
	            	for (int i = 0; i < problem.positions.size(); i++){
	            		if (problem.positions.get(i).month.equals(problem.months.get(m).code) && BalanceCategoryStandardRoute.isBalanceCategoryInRoute(problem.balanceCategories.get(b).code, problem.positions.get(i).route, problem.balanceCategoryStandardRoutes)){
	            			for (int k = 0; k < problem.deckCategories.size(); k++){
	            				if (DeckCategoryToDeckCategory.ifPossibleToPut(problem.deckCategories.get(p).code, problem.deckCategories.get(k).code, problem.deckCategoryToDeckCategories)){
	            					monthlyDemandConstrLHS.addTerm(1.0, l[i][b][p][k]);
	            				}
	            			}
	            		}
	            	}
	            	
	            	GRBLinExpr monthlyDemandConstrRHS = new GRBLinExpr();
	            	monthlyDemandConstrRHS.addConstant(Demand.getSQMDemand(problem.balanceCategories.get(b).code, problem.months.get(m).code, problem.deckCategories.get(p).code, problem.demands));
	        				
	        			model.addConstr(monthlyDemandConstrLHS, GRB.EQUAL, monthlyDemandConstrRHS, "Monthly demand [" + problem.months.get(m).code + "][" + problem.balanceCategories.get(b).code + "][" + problem.deckCategories.get(p).code + "]");
	      			}
	      		}
	      	}
	      }
	      
	      
	      // 4.15
	      // Set in definition of z
	      
	      
	      // 4.16
	      // Set in definition of l
	     
	      
	      // 4.17
	      // Limits the flow of cargoes in a capacity group on a voyage to be within the capacity of the vessel sailing it
	      for (int i = 0; i < problem.positions.size(); i++){
	      	for (int k = 0; k < problem.deckCategories.size(); k++){
	      		for (int xi = 0; xi < problem.capacityGroups.size(); xi++){
	          	GRBLinExpr loadOnVoyageConstrLHS = new GRBLinExpr();
	          	GRBLinExpr loadOnVoyageConstrRHS = new GRBLinExpr();
	          	
	          	for (int b = 0; b < problem.balanceCategories.size(); b++){          		
	          		if (CapacityGroupBalanceCategory.isBalanceCategoryInCapacityGroup(problem.balanceCategories.get(b).code, problem.capacityGroups.get(xi).code, problem.capacityGroupBalanceCategories) && BalanceCategoryStandardRoute.isBalanceCategoryInRoute(problem.balanceCategories.get(b).code, problem.positions.get(i).route, problem.balanceCategoryStandardRoutes)){
	          			for (int p = 0; p < problem.deckCategories.size(); p++){
	          				if (DeckCategoryToDeckCategory.ifPossibleToPut(problem.deckCategories.get(p).code, problem.deckCategories.get(k).code, problem.deckCategoryToDeckCategories)){
	          					loadOnVoyageConstrLHS.addTerm(1.0, l[i][b][p][k]);
	          				}
	          			}
	          		}
	          	}
	          	for (int v = 0; v < problem.vessels.size(); v++){
	          		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
	          			loadOnVoyageConstrRHS.addTerm(VesselSetup.getCapacity(problem.vesselSetups, problem.vessels.get(v).code, problem.positions.get(i).route, problem.deckCategories.get(k).code), y_i[v][i]);
	          		}
	          	}
	          	loadOnVoyageConstrRHS.addTerm(VesselSetup.getMaxCapacity(problem.vesselSetups, problem.deckCategories.get(k).code), dummy[i]);

	      			model.addConstr(loadOnVoyageConstrLHS, GRB.LESS_EQUAL, loadOnVoyageConstrRHS, "Flow of cargoes [" + problem.positions.get(i).id + "][" + problem.deckCategories.get(k).code + "][" + problem.capacityGroups.get(xi).code + "]");
	      		}
	      	}
	      }
			
	   /*
	      // Extra constraint to check the effect of joint speed variables

	      for (int v = 0; v < problem.vessels.size(); v++) {
	      	for (int i = 0; i < problem.positions.size(); i++){
	      		for (int s = 0; s < problem.speedProfiles.size(); s++){
		      		if (problem.vesselCanSailPosition(problem.vessels.get(v), problem.positions.get(i))){
		      			GRBLinExpr jointSpeedVariablesLHS = new GRBLinExpr();
		      			jointSpeedVariablesLHS.addTerm(1.0, x[v][i][s]);
		      		
		      			GRBLinExpr jointSpeedVariablesRHS = new GRBLinExpr();
		      			
		      			for (int j = 0; j < problem.positions.size(); j++){
		      				if (problem.vesselCanSailPositionToPosition(problem.vessels.get(v), problem.positions.get(j), problem.positions.get(i))){
		      					jointSpeedVariablesRHS.addTerm(1.0, x_B[v][j][i][s]);
		      					
		      				}
		      				
		      			}
		      			if (problem.vesselCanSailInitialPosition(problem.vessels.get(v), problem.positions.get(i))){
		      			jointSpeedVariablesRHS.addTerm(1.0, x_Bo[v][i][s]);
		      			}
		      			model.addConstr(jointSpeedVariablesLHS, GRB.EQUAL, jointSpeedVariablesRHS, "Equal speed profiles for voyage sailing and ballast sailing[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "][" + problem.speedProfiles.get(s).code + "]");
		      		}
	      		}
	      	}
	      }
	      */
	      this.env = env;
	      this.model = model;
	      

			}catch(GRBException e) {
	      System.out.println("Error code: " + e.getErrorCode() + ". " +
	          e.getMessage());
	      
	    }
		
			
			
		}
		
		
		/*
		 * ------------------------ Method to solve problem to optimality ----------------------
		 */
		
		public void solveExact(int currentInstance){
			
			try{
				
				// Set solver parameters
				int solutionTimeLimit = 10000;		// Set how long solver will run
				int solutionLimit = 0;						// Set to 1 to return first feasible solution
				
			      
				
	      // (Solution limit, 1) = Returns first feasible solution found
	      if (solutionLimit > 0){
	      	this.model.getEnv().set(GRB.IntParam.SolutionLimit, solutionLimit);    
	      }
	      
	      // Set the termination-criteria to when the gap is equal to 0.01
	      //this.model.getEnv().set(GRB.DoubleParam.MIPGap, GlobalConstants.acceptanceCriteria);
	      
	      // Set the termination-criteria to High Quality
	      //this.model.getEnv().set(GRB.DoubleParam.MIPGap, GlobalConstants.qualityGAPs.get(currentInstance));
	      
	      // Set the termination-criteria to Good Quality
	      //this.model.getEnv().set(GRB.DoubleParam.MIPGap, GlobalConstants.goodGAPs.get(currentInstance));
	      
	      // Set time limit when solving
	      this.model.getEnv().set(GRB.DoubleParam.TimeLimit, solutionTimeLimit);
	      
	      
	      
	      // Set MIPFocus ( 1 = find good solutions, 2 = prove optimality, 3 = focus on the bound)
	      this.model.getEnv().set(GRB.IntParam.MIPFocus, 3);

	      
	      
	      // Run model
	      this.model.optimize();
	  
	      // Record ending time for the whole MIP
	   	  this.endTime = System.currentTimeMillis();
	      
	      
				// For debugging purposes:
	      int optimstatus = this.model.get(GRB.IntAttr.Status);
				if (optimstatus == GRB.Status.INFEASIBLE){
					System.out.println("INFEASIBLE");
					this.model.computeIIS();
					this.model.write("mymodel.ilp");
				}
	      
	      

			}catch(GRBException e) {
	      System.out.println("Error code: " + e.getErrorCode() + ". " +
	          e.getMessage());
	      
	    }
		}
	
				
		/*
		 * ------------------------ Method to solve problem to optimality for each time period in RHH ----------------------
		 */
		
		public void solveMIPinRHH(){

			try {
	      
	      // Set MIPFocus ( 1 = find good solutions, 2 = prove optimality, 3 = focus on the bound)
	      this.model.getEnv().set(GRB.IntParam.MIPFocus, 3);
	      
	      // Set the termination-criteria to when the gap is equal to 0.01
	      // this.model.getEnv().set(GRB.DoubleParam.MIPGap, GlobalConstants.acceptanceCriteria);
	      
	      // Set the termination-criteria to a High Schedule Quality
	      //this.model.getEnv().set(GRB.DoubleParam.MIPGap, GlobalConstants.qualityGAPs.get(currentInstance));
	      
	      // Set the termination-criteria to a Good Schedule Quality
	      this.model.getEnv().set(GRB.DoubleParam.MIPGap, GlobalConstants.goodGAPs.get(currentInstance));
	      
	      // Run model
	      this.model.optimize();

	      
	      
				// For debugging purposes:
	      int optimstatus = this.model.get(GRB.IntAttr.Status);
				if (optimstatus == GRB.Status.INFEASIBLE){
					System.out.println("INFEASIBLE");
					this.model.computeIIS();
					this.model.write("mymodel.ilp");
				}
	      
	      

			}catch(GRBException e) {
	      System.out.println("Error code: " + e.getErrorCode() + ". " +
	          e.getMessage());
	      
	    }
		}

		
		
	
		
		
		
		/*
		 * ----------------------- Method to find an initial feasible solution to a problem ----------------------------
		 */
		
		public void findInitialFeasibleSolution(){
			
			try{
				
				// Set solver parameters
				int solutionTimeLimit = 60;			// Set how long solver will run
				int solutionLimit = 1;					// Set to 1 to return first feasible solution
				
			      
				
	      // (Solution limit, 1) = Returns first feasible solution found
	      if (solutionLimit > 0){
	      	this.model.getEnv().set(GRB.IntParam.SolutionLimit, solutionLimit);    
	      }
	      
	      
	      
	      // Set time limit when solving
	      this.model.getEnv().set(GRB.DoubleParam.TimeLimit, solutionTimeLimit);
	      
	      
	      
	      // Set MIPFocus ( 1 = find good solutions, 2 = prove optimality, 3 = focus on the bound)
	      this.model.getEnv().set(GRB.IntParam.MIPFocus, 1);

	      
	      
	      // Run model
	      this.model.optimize();

	      
	      
				// For debugging purposes:
	      int optimstatus = this.model.get(GRB.IntAttr.Status);
				if (optimstatus == GRB.Status.INFEASIBLE){
					System.out.println("INFEASIBLE");
					this.model.computeIIS();
					this.model.write("mymodel.ilp");
				}
	      
	      

			}catch(GRBException e) {
	      System.out.println("Error code: " + e.getErrorCode() + ". " +
	          e.getMessage());
	      

	    }
		}
		
		
		
		/*
		 * ------------------------ Method to solve a fixed linear program with variables read from file ----------------------
		 */
		
		public void solveFixedLP(String fileName){
			
			try{
				
				// Set solver parameters
				int solutionTimeLimit = 3600;		// Set how long solver will run
				int solutionLimit = 0;					// Set to 1 to return first feasible solution
				
			      
				
	      // (Solution limit, 1) = Returns first feasible solution found
	      if (solutionLimit > 0){
	      	this.model.getEnv().set(GRB.IntParam.SolutionLimit, solutionLimit);    
	      }
	      
	      
	      
	      // Set time limit when solving
	      this.model.getEnv().set(GRB.DoubleParam.TimeLimit, solutionTimeLimit);
	      
	      
	      
	      // Set MIPFocus ( 1 = find good solutions, 2 = prove optimality, 3 = focus on the bound)
	      this.model.getEnv().set(GRB.IntParam.MIPFocus, 2);

	      
	      // Read variable values from file and fix them
	      try {
	  			File file = new File(fileName);
	  			FileReader fileReader = new FileReader(file);
	  			BufferedReader bufferedReader = new BufferedReader(fileReader);
	  			StringBuffer stringBuffer = new StringBuffer();
	  			
	  			String line;
	  			while ((line = bufferedReader.readLine()) != null) {
	  				stringBuffer.append(line);
	  				
	  				this.model.getVarByName(line).set(DoubleAttr.UB, 1.0);
	  				this.model.getVarByName(line).set(DoubleAttr.LB, 1.0);

	  				stringBuffer.append("\n");
	  				line = "";
	  			}
	  			fileReader.close();
	  			System.out.println("Contents of file:");
	  			System.out.println(stringBuffer.toString());
	  		} catch (IOException e) {
	  			e.printStackTrace();
	  		}
	      
	      
	      
	      // Run model
	      this.model.optimize();

	      
	      
				// For debugging purposes:
	      int optimstatus = this.model.get(GRB.IntAttr.Status);
				if (optimstatus == GRB.Status.INFEASIBLE){
					System.out.println("INFEASIBLE");
					this.model.computeIIS();
					this.model.write("mymodel.ilp");
				}
	      

			}catch(GRBException e) {
	      System.out.println("Error code: " + e.getErrorCode() + ". " +
	          e.getMessage());
	      
	    }
		}
		
		
		/*
		 * -------------------------------- METHOD TO WRITE MIP RESULT TO FILE ----------------------------------------------
		 */
		
		public void writeMIPtoFile(int currentInstance){
			
			Writer writer = null;
			DecimalFormat formatter = new DecimalFormat("#");
			try {
			    writer = new BufferedWriter(new OutputStreamWriter(
			          new FileOutputStream("../Output/MIP_OUTPUT_" + TestInstances.testInstances.get(currentInstance)+".txt"), "utf-8"));
			    writer.write("OUTPUT FROM SOLVING " + this.problem.instanceName + " WITH MIP");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
			    int dummies = 0;
		    	for (int i = 0; i < problem.positions.size(); i++){
					try {
						if ((this.model.getVarByName("dummy[" + this.problem.positions.get(i).id +"]").get(GRB.DoubleAttr.X)) > 0.5){
						dummies++;
						}
					} catch (GRBException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			    writer.write("TOTAL: ");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write(String.format("%-20s", "MONTHS"));
			    writer.write(String.format("%-30s", "Start time of MIP"));
			    writer.write(String.format("%-20s", "Duration of MIP"));
			    writer.write(String.format("%-35s", "Number of unserviced voyages"));
			    writer.write(String.format("%-20s", "Objective value"));
				writer.write(String.format("%-20s", "Best bound"));
	    		writer.write("\n");
	    		writer.write(String.format("%-20d", this.problem.months.size()-1));
			    		writer.write(String.format("%-30s", (new Date()).toString()));
			    		writer.write(String.format("%-20f", (this.endTime - this.startTime)/1000));
			    		writer.write(String.format("%-35d", dummies));
			    		try {
							writer.write(String.format("%-20s", formatter.format(this.model.get(GRB.DoubleAttr.ObjVal))));
				    		writer.write(String.format("%-20s", formatter.format(this.model.get(GRB.DoubleAttr.ObjBound))));
						} catch (GRBException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("VOYAGES COVERED BY DUMMY VARIABLES:" );
			    writer.write(System.getProperty( "line.separator" ));
			    	for (int i = 0; i < problem.positions.size(); i++){
						try {
							if ((this.model.getVarByName("dummy[" + this.problem.positions.get(i).id +"]").get(GRB.DoubleAttr.X)) > 0.5){
							writer.write("" + this.problem.positions.get(i).id);
							writer.write(System.getProperty( "line.separator" ));
							}
						} catch (GRBException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
					}
				writer.write("TOTAL NUMBER OF UNSERVICED VOYAGES: " + dummies);
				writer.write(System.getProperty( "line.separator" ));
				writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("VOYAGE ASSIGNMENTS: ");
			    writer.write(System.getProperty( "line.separator" ));
					for (int v = 0; v < problem.vessels.size();v++){
				    writer.write("VESSEL " + problem.vessels.get(v).code + " SAILS:");
				    writer.write(System.getProperty( "line.separator" ));
						for (int i = 0; i < problem.positions.size(); i++){
							try {
								if (y_i[v][i] != null && (vesselCanSailPosition(problem.vessels.get(v),problem.positions.get(i), problem)) && y_i[v][i].get(GRB.DoubleAttr.X) > 0.5){
									writer.write("" + this.problem.positions.get(i).id);
									writer.write(System.getProperty( "line.separator" ));
								}
							} catch (GRBException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}
					}
			    
		      	
			} catch (IOException ex) {
			  // report
			} finally {
			   try {writer.close();} catch (Exception ex) {}
			}
			

		}	
		
		/*
		 * -------------------------------- METHOD TO WRITE SRASO-MIP RESULT TO FILE ----------------------------------------------
		 */
		
		public void writeSRASOtoFile(int currentInstance, String currentSpeedProfile, String description, double startTimeSRASO, double endTimeSRASO){
			
			Writer writer = null;
			DecimalFormat formatter = new DecimalFormat("#");
			try {
			    writer = new BufferedWriter(new OutputStreamWriter(
			          new FileOutputStream("../Output/MIP_OUTPUT_SRASO_"+ description +"_" + TestInstances.testInstances.get(currentInstance)+ "_" + currentSpeedProfile +".txt"), "utf-8"));
			    writer.write("OUTPUT FROM SOLVING " + problem.instanceName + " WITH MIP AND ONLY ALLOWING SPEED PROFILE " + currentSpeedProfile + " BEFORE PERFORMING SPEED OPTIMIZATION WITH FIXED LP-SOLUTION");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
			    int dummies = 0;
		    	for (int i = 0; i < problem.positions.size(); i++){
					try {
						if ((this.model.getVarByName("dummy[" + problem.positions.get(i).id +"]").get(GRB.DoubleAttr.X)) > 0.5){
						dummies++;
						}
					} catch (GRBException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			    writer.write("TOTAL: ");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write(String.format("%-20s", "MONTHS"));
			    writer.write(String.format("%-30s", "Start time of MIP"));
			    writer.write(String.format("%-20s", "Duration of MIP"));
			    writer.write(String.format("%-35s", "Number of unserviced voyages"));
			    writer.write(String.format("%-20s", "Objective value"));
				writer.write(String.format("%-20s", "Best bound"));
	    		writer.write("\n");
	    		writer.write(String.format("%-20d", this.problem.months.size()-1));
			    		writer.write(String.format("%-30s", (new Date()).toString()));
			    		writer.write(String.format("%-20f", (endTimeSRASO - startTimeSRASO)/1000));
			    		writer.write(String.format("%-35d", dummies));
			    		try {
							writer.write(String.format("%-20s", formatter.format(this.model.get(GRB.DoubleAttr.ObjVal))));
				    		writer.write(String.format("%-20s", formatter.format(this.model.get(GRB.DoubleAttr.ObjBound))));
						} catch (GRBException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("VOYAGES COVERED BY DUMMY VARIABLES:" );
			    writer.write(System.getProperty( "line.separator" ));
			    	for (int i = 0; i < problem.positions.size(); i++){
						try {
							if ((this.model.getVarByName("dummy[" + this.problem.positions.get(i).id +"]").get(GRB.DoubleAttr.X)) > 0.5){
							writer.write("" + this.problem.positions.get(i).id);
							writer.write(System.getProperty( "line.separator" ));
							}
						} catch (GRBException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
					}
				writer.write("TOTAL NUMBER OF UNSERVICED VOYAGES: " + dummies);
				writer.write(System.getProperty( "line.separator" ));
				writer.write("-------------------------------------------------------");
			    writer.write(System.getProperty( "line.separator" ));
			    writer.write("VOYAGE ASSIGNMENTS: ");
			    writer.write(System.getProperty( "line.separator" ));
					for (int v = 0; v < problem.vessels.size();v++){
				    writer.write("VESSEL " + problem.vessels.get(v).code + " SAILS:");
				    writer.write(System.getProperty( "line.separator" ));
						for (int i = 0; i < problem.positions.size(); i++){
							try {
								if (y_i[v][i] != null && (vesselCanSailPosition(problem.vessels.get(v),problem.positions.get(i), problem)) && y_i[v][i].get(GRB.DoubleAttr.X) > 0.5){
									writer.write("" + this.problem.positions.get(i).id);
									writer.write(System.getProperty( "line.separator" ));
								}
							} catch (GRBException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}
					}
			    
				    // Writing the 
					int numSpeedUsed = 0;
				      for (int s = 0; s < problem.speedProfiles.size(); s++){
				    	  for (int v = 0; v < problem.vessels.size(); v++) {
				    		  for (int i = 0; i < problem.positions.size(); i++){		
											try {
												if (y_i[v][i] != null && (vesselCanSailPosition(problem.vessels.get(v),problem.positions.get(i), problem)) && y_i[v][i].get(GRB.DoubleAttr.X) > 0.5){	
													if (this.model.getVarByName("x[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "][" + problem.speedProfiles.get(s).code  + "]").get(GRB.DoubleAttr.X) > 0.1){
														numSpeedUsed++;
														writer.write(System.getProperty( "line.separator" ));
														writer.write("Sailing voyage "+ this.problem.positions.get(i).id + "with vessel: " + this.problem.vessels.get(v).code + "  speedprofile: " + this.problem.speedProfiles.get(s).code +" is weighted with the value: " + this.model.getVarByName("x[" + problem.vessels.get(v).code + "][" + problem.positions.get(i).id + "][" + problem.speedProfiles.get(s).code  + "]").get(GRB.DoubleAttr.X));
													}
												}
											} catch (GRBException e) {
												// TODO Auto-generated catch block
												e.printStackTrace();
											}
				      		}
				    	  }
				    		  System.out.println("Speedprofil: " + problem.speedProfiles.get(s).code + " blir brukt: " + numSpeedUsed + " Antall ganger");
				    		  numSpeedUsed = 0;
				      	
				      }
		      	
			} catch (IOException ex) {
			  // report
			} finally {
			   try {writer.close();} catch (Exception ex) {}
			}
			

		}	
		/*
		 * ------------------------- Method to evaluate if a vessel can sail a voyage ------------------------------------------------
		 */

		public static boolean vesselCanSailPosition(Vessel v, Position i, Problem problem){
			if (i.endTime + GlobalConstants.maxDaysDelayed >= v.dateOffset){
				return true;
			}
			else{
				return false;
			}
		}
		
		/*
		 * ------------------------- Method to print a solution to file -------------------------
		 */

		void printVariables(int currentInstance) throws GRBException{
			
			Writer writer = null;
		
			try {
				writer = new BufferedWriter(new OutputStreamWriter(
		          new FileOutputStream("SolutionVariables_" + TestInstances.testInstances.get(currentInstance)+".txt"), "utf-8"));
		    
		    for (int i = 0; i < this.problem.positions.size(); i++){
		    		if ((this.model.getVarByName("dummy[" + this.problem.positions.get(i).id +"]").get(GRB.DoubleAttr.X)) > 0.5){
					    writer.write("dummy["+ this.problem.positions.get(i).id +"]");		
					    writer.write(System.getProperty( "line.separator" ));
		    		}
				}
		    
		    for (int v = 0; v < this.problem.vessels.size(); v++){
		    	for (int i = 0; i < this.problem.positions.size(); i++){
		    		if (y_i[v][i] != null && vesselCanSailPosition(problem.vessels.get(v),problem.positions.get(i), problem) && y_i[v][i].get(GRB.DoubleAttr.X) > 0.5){
		    			writer.write("y_i[" + this.problem.vessels.get(v).code + "][" + this.problem.positions.get(i).id + "]");	
		    			writer.write(System.getProperty( "line.separator" ));
		    		}	
		    	}
		    }
		    
		    		
			} catch (IOException ex) {
		  // report
			} finally {
		   try {writer.close();} catch (Exception ex) {}
			}
		}

}
