import gurobi.GRB;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;

/*
 * This class contains methods to read in a workbook, create a Problem object and store the data in it.
 */

public class Problem{	
	String instanceName;
	ArrayList<Month> months;
	ArrayList<Vessel> vessels;
	ArrayList<SpeedProfile> speedProfiles;
	ArrayList<VesselSpeedProfile> vesselSpeedProfiles;
	ArrayList<Port> ports;
	ArrayList<Position> positions;
	ArrayList<StandardRoute> standardRoutes;
	ArrayList<Sailing> sailings;
	ArrayList<SailingBallastStart> sailingBallastStarts;
	ArrayList<SailingBallast> sailingBallasts;
	ArrayList<CapacityGroup> capacityGroups;
	ArrayList<BalanceCategory> balanceCategories;
	ArrayList<CapacityGroupBalanceCategory> capacityGroupBalanceCategories;
	ArrayList<BalanceCategoryStandardRoute> balanceCategoryStandardRoutes;
	ArrayList<Demand> demands;
	ArrayList<DeckCategory> deckCategories;
	ArrayList<VesselSetup> vesselSetups;
	ArrayList<DeckCategoryToDeckCategory> deckCategoryToDeckCategories;
	
	
	public Problem(HSSFWorkbook workbook, String instanceName){
		this.instanceName = instanceName;
		this.months = Month.importMonths(workbook);
		this.vessels = Vessel.importVessels(workbook);
		this.speedProfiles = SpeedProfile.importSpeedProfiles(workbook);
		this.vesselSpeedProfiles = VesselSpeedProfile.importVesselSpeedProfiles(workbook);
		this.ports = Port.importPorts(workbook);
		this.positions = Position.importPositions(workbook);
		this.standardRoutes = StandardRoute.importStandardRoutes(workbook);
		this.sailings = Sailing.importSailings(workbook);
		this.sailingBallastStarts = SailingBallastStart.importSailingBallastStarts(workbook);
		this.sailingBallasts = SailingBallast.importSailingBallasts(workbook);
		this.capacityGroups = CapacityGroup.importCapacityGroups(workbook);
		this.balanceCategories = BalanceCategory.importBalanceCategories(workbook);
		this.capacityGroupBalanceCategories = CapacityGroupBalanceCategory.importCapacityGroupBalanceCategories(workbook);
		this.balanceCategoryStandardRoutes = BalanceCategoryStandardRoute.importBalanceCategoryStandardRoutes(workbook);
		this.demands = Demand.importDemands(workbook);
		this.deckCategories = DeckCategory.importDeckCategories(workbook);
		this.vesselSetups = VesselSetup.importVesselSetups(workbook);
		this.deckCategoryToDeckCategories = DeckCategoryToDeckCategory.importDeckCategoryToDeckCategories(workbook);
	}
	
	
	
  /*
   *  -------------------- Method to evaluate if a vessel can sail a position -------------------------------------------------------------
   */
	
	public boolean vesselCanSailPosition(Vessel v, Position i){
		
		if (i.endTime + GlobalConstants.maxDaysDelayed >= v.dateOffset
				
				&&
				
				Sailing.getShortestSailingDuration(v.code, i.route, this.sailings) >= 0
				){
			return true;
		}else{
			return false;
		}
	}
	
	
  /*
   *  -------------------- Method to evaluate if a vessel can sail a position as its first scheduled position -----------------------------
   */
	
	public boolean vesselCanSailInitialPosition(Vessel v, Position i){		
		if (i.endTime + GlobalConstants.maxDaysDelayed >= v.dateOffset
				+ SailingBallastStart.getShortestSailingBallastStartDuration(v.code, i.route, this.sailingBallastStarts)	
				
				&&
				
				SailingBallastStart.getShortestSailingBallastStartDuration(v.code, i.route, this.sailingBallastStarts) >= 0
				
				&&
				
				Sailing.getShortestSailingDuration(v.code, i.route, this.sailings) >= 0

			){
				return true;
		}else{
				return false;
		}
	}
	
	
  /*
   *  --------------------- Method to evaluate if a vessel can sail two positions in succession -----------------------------
   */
	
	public boolean vesselCanSailPositionToPosition(Vessel v, Position i, Position j){
		if (
				this.vesselCanSailPosition(v,i)
				&&
				this.vesselCanSailPosition(v,j)
				&&
				i != j
				&&
				i.startTime + Sailing.getShortestSailingDuration(v.code, i.route, this.sailings) + SailingBallast.getShortestSailingBallastDuration(v.code, i.route, j.route, this.sailingBallasts) 
				<= 
				j.endTime + GlobalConstants.maxDaysDelayed
				&&
				SailingBallast.getShortestSailingBallastDuration(v.code, i.route, j.route, this.sailingBallasts) >= 0
				){
			return true;
		}else{
			return false;
		}
	}
	
	
}

