import gurobi.*;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import java.util.*;


/*
 * ------- This class contains the Iteration object, which is used for storing information from each ALNS iteration ---------------
 */


public class Iteration {
	public int iterationNumber;
	public int destroyOperator;
	public int repairOperator;
	public int localSearch;
	public int numberOfDummies;
	public double iterationTime;
	public double[][] y_i;
	public double[] dummyPositions;
	public double[] t;
	public double timeConsumption;
	public double startTime;
	public double objectiveValue;
	
	
	
	/*
	 * ------------------------------- Iteration object constructor -------------------------------
	 */
	
	public Iteration(int iterationNumber, double elapsedTime, int maxIterations, Solution solution){
		this.iterationNumber = iterationNumber+1;
		this.iterationTime = elapsedTime;
		this.y_i = new double[solution.problem.vessels.size()][solution.problem.positions.size()];
		this.dummyPositions = new double[solution.problem.positions.size()];
		this.t = new double[solution.problem.positions.size()];
		this.numberOfDummies = 0;
		this.startTime = System.currentTimeMillis();
		this.objectiveValue = Double.POSITIVE_INFINITY;
		
		solution.selectOperators(this);
		
	}
	
	
	
	public Iteration(){
		
	}
	

	
	/*
	 * --------------------- Method to update Iteration object variables from Solution object----------------
	 */
	
	public void addSolution(Solution solution){
		
		int dummies = 0;
		
		// Update arrays of dummy positions and vessel schedules from Solution object
		for (int i = 0; i < solution.problem.positions.size(); i++){
			for (int j = 0; j < solution.dummyPositions.size(); j++){
				if (solution.problem.positions.get(i).id.equals(solution.dummyPositions.get(j).id)){
					this.dummyPositions[i] = 1.0;
					dummies++;
				}
			}
			for (int v = 0; v < solution.vessels.size(); v++){
				for (int j = 0; j < solution.vessels.get(v).assignedPositions.size(); j++){
					if (solution.problem.positions.get(i).id.equals(solution.vessels.get(v).assignedPositions.get(j).id)){
						this.y_i[v][i] = 1.0;
						this.t[i] = solution.vessels.get(v).assignedPositions.get(j).assignedStartTime;
					}		
				}
			}
		}
		
		this.numberOfDummies = dummies;

		
	}
	
	
	
	/*
	 * --------------------- Method to evaluate Solution against previous ones found ----------------
	 */
	
	public void evaluate(ArrayList<Iteration> allIterations, Solution solution, ArrayList<Iteration> bestIterations){		
		
		// Update time spent repairing solution
		this.timeConsumption = (new Date()).getTime() - this.startTime;
		//System.out.println("TIME SPENT DESTROYING AND REPAIRING: " + this.timeConsumption);
		//System.out.println("BEST ITERATIONS SIZE: " + bestIterations.size());
		
		// Check if this iteration is the best yet, with respect to the number of dummies
		if (bestIterations.size() > 0){
			
			if (bestIterations.get(0).numberOfDummies > this.numberOfDummies){
				
				// This is the best iteration yet. We clear the list of best iterations, add this, and award scores.
				//System.out.println("THIS ITERATION IS THE BEST YET. NUMBER OF DUMMIES: " + this.numberOfDummies);
				bestIterations.clear();
				bestIterations.add(this);
				
				// Reward for obtaining a better solution with less dummies
				int score = (int)((double)GlobalConstants.scoreRewardRemovedDummy / ((double)this.timeConsumption/1000));
				if (score > GlobalConstants.scoreRewardRemovedDummy){
					score = GlobalConstants.scoreRewardRemovedDummy;
				}

				solution.scores[(this.destroyOperator)-1][(this.repairOperator)-1] += score;
				//System.out.println("SCORE GIVEN: " + score);
							
				
				
			}else if (bestIterations.get(0).numberOfDummies == this.numberOfDummies){
				
				// The same amount of dummies is obtained, and the solutions must be compared. If they don't have the same schedule, add the current iteration to the list of best iterations
				boolean newSchedule = this.compareSchedules(bestIterations, solution);
				
				/*
				System.out.println("THE DUMMIES IN THIS SOLUTION ARE: ");
				for (int i = 0; i < solution.dummyPositions.size(); i++){
					System.out.println("[" + solution.dummyPositions.get(i).id + "]");
				}
				
				System.out.println("NEW SCHEDULE? " + newSchedule);
				*/
				
				if (newSchedule){
					// We have found a new possible schedule with the same amount of dummies. Add iteration to list of best iterations.
					//System.out.println("THIS ITERATION PROVIDED A NEW POSSIBLE SCHEDULE. NUMBER OF DUMMIES: " + this.numberOfDummies);
					bestIterations.add(this);
					
					// Award scores
					int score = (int)((double)GlobalConstants.scoreRewardImproved / ((double)this.timeConsumption/1000));
					if (score > GlobalConstants.scoreRewardImproved){
						score = GlobalConstants.scoreRewardImproved;
					}
					
					solution.scores[(this.destroyOperator)-1][(this.repairOperator)-1] += score;
					//System.out.println("SCORE GIVEN: " + score);
				}else{
					// New schedule not found.
					//System.out.println("THIS ITERATION GAVE US A SCHEDULE ALREADY FOUND. NUMBER OF DUMMIES: " + this.numberOfDummies);
				}
			
	
				
			}else{
				
				// This is not one of the best iterations yet (more dummies). Evaluate whether it should be accepted as a part of simulated annealing or not.
				double delta = 0;
				
				// Apply Simulated Annealing to evaluate whether solution should be accepted or not
				if (bestIterations.get(0).numberOfDummies > 0){
					delta = (double)(this.numberOfDummies - bestIterations.get(0).numberOfDummies)/((double)bestIterations.get(0).numberOfDummies) ;
				}else{
					delta = (double)(this.numberOfDummies - bestIterations.get(0).numberOfDummies)/(1) ;
				}
				
				//double T = solution.maxTime - this.iterationTime;
				//double temperatureFactor = (T / solution.maxTime); 	//* GlobalConstants.annealingFactor;	// Will give a percentage of time left of search, times the user-defined annealing factor
				double acceptanceCriterion = Math.exp((-delta)/(solution.temperature));
				double acceptanceProbability = (double)(Math.random());

				if (acceptanceProbability > acceptanceCriterion){
					
					// Do not accept solution, do not give any scores, and go back to Solution from best iteration
					int iter = solution.chooseFromListOfBestIterations(bestIterations);
					
					//System.out.println("SOLUTION FOUND: " + this.numberOfDummies);
					//System.out.println("BEST SOLUTION YET: " + bestIterations.get(0).numberOfDummies);
					//System.out.println("NUMBER OF BEST ITERATIONS WITH SAME # OF DUMMIES: " + bestIterations.size());
					//System.out.println("JUMPED BACK TO BEST SOLUTION AT ITERATION: " + iter);
					
					// Update Iteration object values
					//this.dummyPositions = allIterations.get(bestIterationPosition).dummyPositions;
					//this.y_i = allIterations.get(bestIterationPosition).y_i;
					//this.t = allIterations.get(bestIterationPosition).t;
					
				}else{
					
					// Accept solution, reward score points, and leave Solution object untouched
					int score = (int)((double)GlobalConstants.scoreRewardAccepted / ((double)this.timeConsumption/1000));
					if (score > GlobalConstants.scoreRewardAccepted){
						score = GlobalConstants.scoreRewardAccepted;
					}
					
					solution.scores[(this.destroyOperator)-1][(this.repairOperator)-1] += score;
					//System.out.println("SOLUTION FOUND: " + this.numberOfDummies);
					//System.out.println("BEST SOLUTION YET: " + bestIterations.get(0).numberOfDummies);
					//System.out.println("WORSE SOLUTION ACCEPTED AS A PART OF SIMULATED ANNEALING: " + (this.iterationNumber));
					//System.out.println("SCORE GIVEN: " + score);
				}
			}
			
			
		}else{
			
			// This has to be the first iteration
			//System.out.println("THIS ITERATION IS THE FIRST ONE. NUMBER OF DUMMIES: " + this.numberOfDummies);
			bestIterations.add(this);
			
			// Reward for obtaining a better solution with less dummies
			/*
			int score = (int)((double)GlobalConstants.scoreRewardRemovedDummy / ((double)this.timeConsumption/1000));
			if (score > GlobalConstants.scoreRewardRemovedDummy){
				score = GlobalConstants.scoreRewardRemovedDummy;
			}
				
			solution.scores[(this.destroyOperator)-1][(this.repairOperator)-1] += score;
			System.out.println("SCORE GIVEN: " + (int)((double)GlobalConstants.scoreRewardRemovedDummy / ((double)this.timeConsumption/1000)));
				*/
		}
		
		
		// Cool down Solution 
		//System.out.println("OLD TEMPERATURE: " + solution.temperature);
		solution.temperature = solution.temperature * GlobalConstants.coolingFactor;
		//System.out.println("NEW TEMPERATURE: " + solution.temperature);
			
			
		// UPDATE OPERATOR SCORES
		solution.updateWeights(allIterations);
		//System.out.println("----------------------------------");
		System.out.println("----------------------------------");

		
	}
	
	
	
	
	/*
	 * ----------------------- Method to compare the solution of an iteration to the schedules in the list of best iterations ----------------------
	 */
	
	public boolean compareSchedules(ArrayList<Iteration> bestIterations, Solution solution){
		
		
		for (int k = 0; k < bestIterations.size(); k++){
			boolean scheduleMatch = true;
			
			//System.out.println("ITERATION " + k + " HAS THE FOLLOWING DUMMIES:");
			
			for (int i = 0; i < solution.problem.positions.size(); i++){
				
				
				//TEST:
				//if (bestIterations.get(k).dummyPositions[i] > 0.5){
				//	System.out.println("[" + solution.problem.positions.get(i).id + "]");
				//}
				
				
				if ((bestIterations.get(k).dummyPositions[i] > 0.5) && !(this.dummyPositions[i] > 0.5)){
					scheduleMatch = false;
				}			
			}
			
			if (scheduleMatch){
				return false;
			}
		}

		return true;
	}
	
	
	
	
	/*
	 * --------------------- Method to update a Solution object from an Iteration Object ------------
	 */
	
	public void getSolutionFromIteration(Solution solution){
		
		// Clear Solution
		solution.dummyPositions.clear();
		for (int v = 0; v < solution.vessels.size(); v++){
			solution.vessels.get(v).assignedPositions.clear();
		}
		
		// Rebuild Solution
		for (int i = 0; i < solution.problem.positions.size(); i++){
			
			// Update dummy positions
			if (this.dummyPositions[i] > 0.5){
				solution.dummyPositions.add(solution.problem.positions.get(i));
			}
			
			// BØR KANSKJE PASSE PÅ REBYGGINGEN HER, MTP STARTTIDSPUNKTER OG DET KAN BLI FEIL REKKEFØLGE. SORTERE PÅ ASSIGNED START TIME
			
			// Update vessel positions
			for (int v = 0; v < solution.problem.vessels.size(); v++){
				if (this.y_i[v][i] > 0.5){
					solution.vessels.get(v).assignedPositions.add(solution.problem.positions.get(i));
					solution.vessels.get(v).assignedPositions.get(solution.vessels.get(v).assignedPositions.size()-1).assignedStartTime = this.t[i];
				}
			}
		}
	}

	
	
	/*
	 * ------------ Method to get the number of dummy variables in an iteration --------------
	 */
	
	public int getNumberOfDummyVariables(){
		int numberOfDummies = 0;
		
		for (int i = 0; i < this.dummyPositions.length; i++){
			if (this.dummyPositions[i] > 0.5){
				numberOfDummies++;
			}
		}
		return numberOfDummies;
	}
}
