import gurobi.GRB;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;

/*
 * This class creates a Solution object, which is used in the ALNS iterations.
 */

public class Solution {
	ArrayList<Position> unassignedPositions;
	ArrayList<Position> dummyPositions;
	ArrayList<Vessel> vessels;
	Problem problem;
	double maxTime;
	double temperature;
	int[][] scores;
	double[][][] weights;

	
	
	// Constructor of an initial feasible Solution object from a Problem
	public Solution(Problem problem){
		
		System.out.println("---------------------------------------------");
		System.out.println("CONSTRUCTING INITIAL SOLUTION...");

			
			// Initialize a list of unassigned positions, and set all positions as "unassigned"
			this.unassignedPositions = new ArrayList<Position>();
			for (int i = 0; i < problem.positions.size(); i++){
				this.unassignedPositions.add(problem.positions.get(i));
			}
			
			// Create a list of dummyPositions
			this.dummyPositions = new ArrayList<Position>();
			
			
			// Create a list of vessels, and for each vessel create an empty list of positions (the vessel schedule)
			this.vessels = new ArrayList<Vessel>();
			for (int v = 0; v < problem.vessels.size(); v++){
				this.vessels.add(problem.vessels.get(v));
				this.vessels.get(v).assignedPositions = new ArrayList<Position>();
			}
			
			
			// Store problem instance in Solution object
			this.problem = problem;
			
			
			// Initiate operator scores and weights
			this.scores = new int [GlobalConstants.numberOfDestroyOperators][GlobalConstants.numberOfRepairOperators];
			this.weights = new double [GlobalConstants.numberOfDestroyOperators][GlobalConstants.numberOfRepairOperators][GlobalConstants.maximumNumberOfIterations / GlobalConstants.segmentSize];
			
			for (int d = 0; d < GlobalConstants.numberOfDestroyOperators; d++){
				for (int r = 0; r < GlobalConstants.numberOfRepairOperators; r++){
					this.scores[d][r] = 0;
					
					for (int j = 0; j < (GlobalConstants.maximumNumberOfIterations / GlobalConstants.segmentSize); j++){
						this.weights[d][r][j] = 1.00000;
					}
				}
			}
			
			
			// Set Solution attributes
			this.maxTime = GlobalConstants.maxTimer;
			this.temperature = GlobalConstants.startTemperature;
			
			
			// Find initial solution
			Collections.shuffle(this.unassignedPositions);
			RepairOperator.regret_k_Insertion(this, 4);
			System.out.println("NUMBER OF DUMMIES IN INITIAL SOLUTION: " + this.dummyPositions.size());
			RepairOperator.localSearch(this);
			System.out.println("NUMBER OF DUMMIES IN AFTER REASSIGN: " + this.dummyPositions.size());
			System.out.println("---------------------------------------------");

	}
	
	
	
	
	/*
	 * -------------------- Method to swap a pair of positions -----------------------------------------
	 */
	
	public void swap (Position position1, Position position2){
		
		int sequence1 = - 1;
		int sequence2 = - 1;
		int vessel1 = - 1;
		int vessel2 = - 1;	
		
		// First, find vessels and indexes of the positions
		for (int v = 0; v < this.vessels.size(); v++){
			for (int i = 0; i < this.vessels.get(v).assignedPositions.size(); i++){
				if (this.vessels.get(v).assignedPositions.get(i) == position1){
					vessel1 = v;
					sequence1 = i;
				}else if (this.vessels.get(v).assignedPositions.get(i) == position2){
					vessel2 = v;
					sequence2 = i;				
				}		
			}
		}
		
		//TEST
		//System.out.println("OLD SCHEDULES:");
		//this.vessels.get(vessel1).printSchedule();
		//this.vessels.get(vessel2).printSchedule();
		
		
		// Swap positions
		this.vessels.get(vessel1).assignedPositions.remove(position1);
		this.vessels.get(vessel2).assignedPositions.remove(position2);
		this.vessels.get(vessel1).assignedPositions.add(sequence1, position2);
		this.vessels.get(vessel2).assignedPositions.add(sequence2, position1);
		
		// Assign new start times
		this.vessels.get(vessel1).assignStartTimes(this);
		this.vessels.get(vessel2).assignStartTimes(this);
	
		
		//TEST
		//System.out.println("NEW SCHEDULES:");
		//this.vessels.get(vessel1).printSchedule();
		//this.vessels.get(vessel2).printSchedule();
		
	}
	
	
	
	
	/*
	 * -------------------- Method to find a possible insertion for a position -------------------------
	 */
	
	public boolean possibleInsertionFoundFor(Position j){
			
		// Search all vessels for a possible insertion for position j
		int v = 0;
		while (v < this.vessels.size()){
					
			// If vessel schedule is non-empty, check where the position may be inserted
			if (!(this.vessels.get(v).assignedPositions.isEmpty())){
							
				// Check if the voyage can be inserted in the first position
				if (j.canBeAddedInSequence(this.vessels.get(v), 0, this)){		
					this.vessels.get(v).assignedPositions.add(0, j);
					this.vessels.get(v).assignStartTimes(this);
					
					//TEST
					//System.out.println("POSSIBLE INSERTION FOUND FOR POSITION " + j.id + " AT VESSEL " + this.vessels.get(v).code);
					
					return true;			
				}
				
				// Check if the voyage can be inserted in between positions in the vessel schedule
				for (int i = 1; i < this.vessels.get(v).assignedPositions.size(); i++){
					if (j.canBeAddedInSequence(this.vessels.get(v), i, this)){
						this.vessels.get(v).assignedPositions.add(i, j);
						this.vessels.get(v).assignStartTimes(this);
						
						//TEST
						//System.out.println("POSSIBLE INSERTION FOUND FOR POSITION " + j.id + " AT VESSEL " + this.vessels.get(v).code);
						
						return true;
					}
				}
						
				// Check if the voyage can be inserted at the end of the list
				if (j.canBeAddedInSequence(this.vessels.get(v), this.vessels.get(v).assignedPositions.size(), this)){
					this.vessels.get(v).assignedPositions.add(this.vessels.get(v).assignedPositions.size(), j);
					this.vessels.get(v).assignStartTimes(this);
					
					//TEST
					//System.out.println("POSSIBLE INSERTION FOUND FOR POSITION " + j.id + " AT VESSEL " + this.vessels.get(v).code);
					
					return true;
				}
				
			
			// If a vessel's schedule is empty, check if the voyage can be served
			}else{
				if (this.problem.vesselCanSailInitialPosition(this.vessels.get(v), j)){
					this.vessels.get(v).assignedPositions.add(0, j);
					this.vessels.get(v).assignStartTimes(this);	
					
					//TEST
					//System.out.println("POSSIBLE INSERTION FOUND FOR POSITION " + j.id + " AT VESSEL " + this.vessels.get(v).code);
					
					return true;
				}
			}
				
			v++;
		} // End while v

		//TEST
		//System.out.println("NO POSSIBLE INSERTION FOUND FOR POSITION " + j.id);
		
		
		// No insertions are found
		return false;
			
		
	}
	
	
	
	
	/*
	 *  ---------------- Method to select destroy and repair operators ------------------
	 */
	
	public void selectOperators(Iteration iteration){
		
		// Choose random combination of destroy and repair operators based on weights
		double random = (double )(Math.random());
		double sumOfWeights = 0;
		double sumOfThresholds = 0;
		boolean selectionDone = false;
		int dd = 0;
		int rr = 0;
		
		
		// Find the sum of weights
		for (int d = 0; d < GlobalConstants.numberOfDestroyOperators; d++){
			for (int r = 0; r < GlobalConstants.numberOfRepairOperators; r++){
				sumOfWeights += this.weights[d][r][(int)(iteration.iterationNumber - 1) / (GlobalConstants.segmentSize)];
			}
		}
		
		//TEST
		//System.out.println("RANDOM: " + random);

		
		// Choose operators based on thresholds defined by weights
		while (dd < GlobalConstants.numberOfDestroyOperators && !selectionDone){
			rr = 0;
			while (rr < GlobalConstants.numberOfRepairOperators && !selectionDone){
				sumOfThresholds += (this.weights[dd][rr][(int)(iteration.iterationNumber - 1) / (GlobalConstants.segmentSize)] / sumOfWeights);
				
				//TEST
				//System.out.println("SUM OF THRESHOLDS: " + sumOfThresholds);
				
				if (random <= sumOfThresholds){
					//System.out.println("DESTROY: " + (dd + 1));
					//System.out.println("REPAIR: " + (rr + 1));

					iteration.destroyOperator = (dd + 1);
					iteration.repairOperator = (rr + 1);
					selectionDone = true;
				}
				rr++;
			}
			dd++;
		}
		
		// Choose local search operator
		int randomlocalSearch = (int)(Math.random() * GlobalConstants.numberOfLocalSearchNeighborhoods + 1);
		//System.out.println("LOCAL SEARCH: " + randomlocalSearch);
		iteration.localSearch = randomlocalSearch;
				
	}
		
		//System.out.println("SCORES: " + solution.destroyScores[0] + " " + solution.destroyScores[1] + " " + solution.destroyScores[2] + " " + solution.destroyScores[3]  + " ");
		//System.out.println("THRESHOLDS: " + threshold1 + " " + threshold2 + " " + threshold3 + " ");
	
	
	
	
	public void createSolutionObject(Mip mip){
		try{
			
		// Initialize an empty list of unassigned positions
		this.unassignedPositions = new ArrayList<Position>();

					
		// Create a list of dummyPositions from the input mip.model
		this.dummyPositions = new ArrayList<Position>();
		for (int i = 0; i < mip.problem.positions.size(); i++){
			if(mip.model.getVarByName("dummy[" + mip.problem.positions.get(i).id + "]").get(GRB.DoubleAttr.X) > 0.5){
				this.dummyPositions.add(mip.problem.positions.get(i));
			}
		}
		
		// Create a list of vessels, and for each vessel create an empty list of positions (the vessel schedule)
		this.vessels = new ArrayList<Vessel>();
		for (int v = 0; v < mip.problem.vessels.size(); v++){
			this.vessels.add(mip.problem.vessels.get(v));
			this.vessels.get(v).assignedPositions = new ArrayList<Position>();
		}
		
		// Iterate all positions. If a positions is assigned to a vessel, add that position to the vessel schedule
		for (int i = 0; i < mip.problem.positions.size(); i++){
			for (int v = 0; v < mip.problem.vessels.size(); v++){
				if (mip.problem.vesselCanSailPosition(mip.problem.vessels.get(v), mip.problem.positions.get(i))){
					if (mip.model.getVarByName("y_i[" + mip.problem.vessels.get(v).code + "][" + mip.problem.positions.get(i).id + "]").get(GRB.DoubleAttr.X) > 0.5){
						this.vessels.get(v).assignedPositions.add(mip.problem.positions.get(i));
					}
				}
			}
		}	
		
		
		// Store problem instance in Solution object
		this.problem = mip.problem;
		
		
		// Initiate operator scores and weights
		this.scores = new int [GlobalConstants.numberOfDestroyOperators][GlobalConstants.numberOfRepairOperators];
		this.weights = new double [GlobalConstants.numberOfDestroyOperators][GlobalConstants.numberOfRepairOperators][GlobalConstants.maximumNumberOfIterations / GlobalConstants.segmentSize];
		
		for (int d = 0; d < GlobalConstants.numberOfDestroyOperators; d++){
			for (int r = 0; r < GlobalConstants.numberOfRepairOperators; r++){
				this.scores[d][r] = 0;
				
				for (int j = 0; j < (GlobalConstants.maximumNumberOfIterations / GlobalConstants.segmentSize); j++){
					this.weights[d][r][j] = 1.00000;
				}
			}

		}
		
		
		}catch(GRBException e) {
      System.out.println("Error code: " + e.getErrorCode() + ". " +
      e.getMessage());
      e.printStackTrace();
    }
	}
	
	
	
	/*
	 * ------------------------------ Method to update scores ----------------------------------------
	 */
	
	public void updateWeights(ArrayList<Iteration> allIterations){
		
		
		if (allIterations.size() % GlobalConstants.segmentSize != 0){
			/* PROCEED SEARCH
			// Update scores
			if (allIterations.get(allIterations.size()-1).improvedSolution){
				this.destroyScores[allIterations.get(allIterations.size()-1).destroyOperator-1] = this.destroyScores[allIterations.get(allIterations.size()-1).destroyOperator-1] + GlobalConstants.scoreRewardImproved;
			}
			*/
		}else{
			if (allIterations.size() < GlobalConstants.maximumNumberOfIterations){ 		// To protect us from the last iteration
				
				// For every 50th iteration, reset scores and calculate smoothened weights to be used in next segment (next M iterations)
				for (int d = 0; d < GlobalConstants.numberOfDestroyOperators; d++){
					for (int r = 0; r < GlobalConstants.numberOfRepairOperators; r++){
						
						// Smooth weights
						this.weights[d][r][(int) (allIterations.size()/GlobalConstants.segmentSize)] = GlobalConstants.reactionFactor * (this.scores[d][r] / this.getNumberOfOperatorCalls(allIterations, (d+1), (r+1))) + (1 - GlobalConstants.reactionFactor) * this.weights[d][r][(int)(allIterations.size()/GlobalConstants.segmentSize) - 1];
					
						// Reset scores
						this.scores[d][r] = 0;
					}
				}
				
			// Reset weights every 1000 iterations
				if(allIterations.size() % GlobalConstants.resetWeightsEvery == 0){
					for (int d = 0; d < GlobalConstants.numberOfDestroyOperators; d++){		
						for (int r = 0; r < GlobalConstants.numberOfRepairOperators; r++){
							this.weights[d][r][(int) (allIterations.size()/GlobalConstants.segmentSize)] = 1.000;
						}
					}
				}	
			}
		}
	}
	
	
	
	/*
	 * ------------------------------ Method to get destroy operator weights ----------------------------------------
	 */
	
	public double getDestroyOperatorWeights(int destroyOperator, int repairOperator, ArrayList<Iteration> allIterations){
		return this.weights[destroyOperator-1][repairOperator-1][(int)(allIterations.size() -1)/ GlobalConstants.segmentSize];
	}
	
	
	
	/*
	 * ------------------------------ Method to get the number a pair of operators are called ----------------------------------------
	 */
	
	public int getNumberOfOperatorCalls(ArrayList<Iteration> allIterations, int destroyOperator, int repairOperator){
		int n = 0;
		for (int k = 0; k < GlobalConstants.segmentSize; k++){
			if (allIterations.get(allIterations.size()-1-k).destroyOperator == destroyOperator 
					&& 
					allIterations.get(allIterations.size()-1-k).repairOperator == repairOperator){
				n++;
			}
		}
		if (n < 1){
			n = 1;
		}
		
		return n;
	}
	
	
	
	/*
	 * ------------------------------ Method to update Solution object from LP-solver ----------------------------------------
	 */
	public void updateSolutionObject(Mip mip){
		try{
			
			
		// Update positions covered by dummies
		this.dummyPositions.clear();
		for (int i = 0; i < mip.problem.positions.size(); i++){
			if (mip.model.getVarByName("dummy[" + mip.problem.positions.get(i).id + "]").get(GRB.DoubleAttr.X) > 0.5){
				this.dummyPositions.add(mip.problem.positions.get(i));
			}
		}
		
		// Clear assigned positions for each vessel
		for (int v = 0; v < mip.problem.vessels.size(); v++){
			this.vessels.get(v).assignedPositions.clear();
		}
		
		// Iterate all positions. If a positions is assigned to a vessel v, add that position to the vessel schedule
		for (int i = 0; i < mip.problem.positions.size(); i++){
			for (int v = 0; v < mip.problem.vessels.size(); v++){
				if (mip.problem.vesselCanSailPosition(mip.problem.vessels.get(v), mip.problem.positions.get(i))){
					if (mip.model.getVarByName("y_i[" + mip.problem.vessels.get(v).code + "][" + mip.problem.positions.get(i).id + "]").get(GRB.DoubleAttr.X) > 0.5){
						this.vessels.get(v).assignedPositions.add(mip.problem.positions.get(i));
						this.vessels.get(v).assignedPositions.get(this.vessels.get(v).assignedPositions.size()-1).assignedStartTime = mip.model.getVarByName("t[" + problem.positions.get(i).id + "]").get(GRB.DoubleAttr.X);
					}
				}
			}
		}
		
		// Before destroying, sort vessel schedules and add randomness to search
		for (int v = 0; v < this.vessels.size(); v++){
			Collections.sort(this.vessels.get(v).assignedPositions, Position.PositionAssignedStartTimeComparator);
			this.vessels.get(v).assignStartTimes(this);
		}
		
		/*
		// Assign start times in the vessel schedules
		for (int v = 0; v < this.vessels.size(); v++){
		}
		*/
		
		// Reset model
		mip.model.reset();
		
		}catch(GRBException e) {
      System.out.println("Error code: " + e.getErrorCode() + ". " +
      e.getMessage());
      e.printStackTrace();
    }
	}
	
	
	
	/*
	 * -------------------- Method to rebuild Solution from a randomly chosen Iteration from the list of best iterations ---------------------------
	 */
	
	public int chooseFromListOfBestIterations(ArrayList<Iteration> bestIterations){
		
		// Choose a random iteration
		int randomNum = (int)(Math.random()*(bestIterations.size()));
		bestIterations.get(randomNum).getSolutionFromIteration(this);
		return bestIterations.get(randomNum).iterationNumber;
	}
	
	

	/*
	 * ------------------------- Method to evaluate if a vessel can sail a voyage ------------------------------------------------
	 */

	public static boolean vesselCanSailPosition(Vessel v, Position i, Problem problem){
		if (i.endTime + GlobalConstants.maxDaysDelayed >= v.dateOffset){
			return true;
		}
		else{
			return false;
		}
	}
	/*
	 * 
	 * 
	 * ------------------------- Method to evaluate if a vessel can sail a voyage as an initial position -------------------------
	 */

	public static boolean vesselCanSailInitialPosition(Vessel v, Position i, Problem problem){
		if (i.endTime + GlobalConstants.maxDaysDelayed >= v.dateOffset 
				/*+ SailingBallastStart.getShortestSailingBallastStartDuration(v.code, i.route, problem.sailingBallastStarts)*/){
			return true;
		}
		else{
			return false;
		}
	}
	
	
	
	/*
	 * ------------------------- Method to evaluate if a vessel can sail from a voyage to another voyage -------------------------
	 */

	public static boolean vesselCanSailPositionToPosition(Vessel v, Position i, Position j, Problem problem){
		
		if (
				i.endTime + GlobalConstants.maxDaysDelayed >= v.dateOffset
				&&
				i != j
				&&
				i.startTime + Sailing.getShortestSailingDuration(v.code, i.route, problem.sailings) + SailingBallast.getShortestSailingBallastDuration(v.code, i.route, j.route, problem.sailingBallasts) <= j.endTime + GlobalConstants.maxDaysDelayed
				&&
				j.endTime + GlobalConstants.maxDaysDelayed >= v.dateOffset
				){
			return true;
		}
		else{
			return false;
		}
	}
	
	
	/*
	 * ------------------------- Method to print a solution to file -------------------------
	 */

	public void printToTextFile(){
		
		Writer writer = null;
	
		try {
			writer = new BufferedWriter(new OutputStreamWriter(
	          new FileOutputStream("SOLUTION_INFEASIBLE"+".txt"), "utf-8"));
	    
	    writer.write("dummy_i_sol: [");
	    for (int i = 0; i < this.dummyPositions.size(); i++){
			    writer.write("(" + this.dummyPositions.get(i).id + ") " + 1 + " ");				
			}
	    writer.write("]");
	    
	    writer.write(System.getProperty( "line.separator" ));
	    
	    writer.write("y_vi_sol: [");
	    for (int v = 0; v < this.vessels.size(); v++){
	    	for (int i = 0; i < this.vessels.get(v).assignedPositions.size(); i++){
			    writer.write("(" + this.vessels.get(v).code + " " + this.vessels.get(v).assignedPositions.get(i).id + ") " + 1 + " ");				
	    	}
	    }
	    writer.write("]");
	    
	    		
		} catch (IOException ex) {
	  // report
		} finally {
	   try {writer.close();} catch (Exception ex) {}
		}
	}
	
	
	
	/*
	 * ------------------------- Method to print a solution to file -------------------------
	 */

	public void printVariables(){
		
		Writer writer = null;
	
		try {
			writer = new BufferedWriter(new OutputStreamWriter(
	          new FileOutputStream("VARIABLES_INFEASIBLE"+".txt"), "utf-8"));
	    
	    for (int i = 0; i < this.dummyPositions.size(); i++){
			    writer.write("dummy[" + this.dummyPositions.get(i).id + "]");		
			    writer.write(System.getProperty( "line.separator" ));
			}
	    
	    for (int v = 0; v < this.vessels.size(); v++){
	    	for (int i = 0; i < this.vessels.get(v).assignedPositions.size(); i++){
			    writer.write("y_i[" + this.vessels.get(v).code + "][" + this.vessels.get(v).assignedPositions.get(i).id + "]");	
			    writer.write(System.getProperty( "line.separator" ));
	    	}
	    }
	    
	    		
		} catch (IOException ex) {
	  // report
		} finally {
	   try {writer.close();} catch (Exception ex) {}
		}
	}
	
	
	

	/*
	 * ------------------------- Method to print a destroyed solution to file -------------------------
	 */

	public void printVariablesDestroyed(){
		
		Writer writer = null;
	
		try {
			writer = new BufferedWriter(new OutputStreamWriter(
	          new FileOutputStream("VARIABLES_DESTROYED_INFEASIBLE"+".txt"), "utf-8"));
	    
	    for (int i = 0; i < this.dummyPositions.size(); i++){
			    writer.write("dummy[" + this.dummyPositions.get(i).id + "]");		
			    writer.write(System.getProperty( "line.separator" ));
			}
	    
	    for (int i = 0; i < this.unassignedPositions.size(); i++){
	    	writer.write("dummy[" + this.unassignedPositions.get(i).id + "]");		
		    writer.write(System.getProperty( "line.separator" ));
	    }
	    
	    for (int v = 0; v < this.vessels.size(); v++){
	    	for (int i = 0; i < this.vessels.get(v).assignedPositions.size(); i++){
			    writer.write("y_i[" + this.vessels.get(v).code + "][" + this.vessels.get(v).assignedPositions.get(i).id + "]");	
			    writer.write(System.getProperty( "line.separator" ));
	    	}
	    }
	    
	    		
		} catch (IOException ex) {
	  // report
		} finally {
	   try {writer.close();} catch (Exception ex) {}
		}
	}

}
